module Formosa
  # Holo language (Southern Min) module
  module Holo
    # Syllable type supported by Formosa    
    module SyllableType
      POJ = 0
      TL = 1
    end
  
    # Holo syllable utiliy
    module SyllableUtility
      # Compose Holo syllable from the query form ("ASCII form")
      #
      # * input_type: can be either SyllableType::POJ or Syllable::TL
      # * output_type: same as above
      # * syllable: the query form of the syllable, such as "goa2", "tai5"
      #
      # Conversion is done automatically when input and output types are different
      def self.compose_syllable(input_type, output_type, syllable, force_poj_style_output = false)
        NativeSyllableComposer.compose(input_type, output_type, syllable, force_poj_style_output ? 1 : 0)
      end

      TONE_SAMPLE = ["a", "a", "á", "à", "a", "â", "ǎ", "ā", "a̍", "a̋"]
      LOOKUP_TABLE = {}

      9.times do |index|
        t = TONE_SAMPLE[index]
        c = t.chars.decompose

        if c.size > 1
          tone_char = [c[1]].pack("U")
          LOOKUP_TABLE[tone_char] = index
        end
      end    

      # Converts a long text of composed syllables into the query form
      def self.convert_text_into_query_form(input_type, input_string)
        output_string = ""

        input_string.split(/([^A-Za-z\x7f-\xff])/).each do |str|
          qform = self.convert_syllable_into_query_form(input_type, str)
          output_string += qform
        end

        output_string
      end
      
      # Converts a composed syllable (the "composed form") into the query form.
      # For example, "guá" is converted to "gua2"
      def self.convert_syllable_into_query_form(input_type, string)
        c = string.chars

        # decomposed code points
        dcps = c.decompose

        loudest_tone = 0

        composed = []

        dcps.size.times do |index|
          chr = [dcps[index]].pack("U")

          if t = LOOKUP_TABLE[chr]
            loudest_tone = t
          else      
            composed << dcps[index]
          end
        end

        result = composed.pack("U*") + (loudest_tone > 0 ? loudest_tone.to_s : "")
      
        if input_type == SyllableType::TL
          result.gsub!(/O\315\230/, "OO")
          result.gsub!(/o\315\230/, "oo")
        else
          result.gsub!(/O\315\230/, "OU")
          result.gsub!(/o\315\230/, "ou")
        end
      
        result.gsub!(/ⁿ/, "nn")
        result
      end
    
    end
  end
end